'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var errors = require('../errors.cjs');
var createListing = require('./createListing.cjs');
var executeSale = require('./executeSale.cjs');
var Bid = require('../models/Bid.cjs');
var Operation = require('../../../types/Operation.cjs');
var PublicKey = require('../../../types/PublicKey.cjs');
var DateTime = require('../../../types/DateTime.cjs');
var pdas = require('../../tokenModule/pdas.cjs');
var TransactionBuilder = require('../../../utils/TransactionBuilder.cjs');

// Operation
// -----------------

const Key = 'DirectSellOperation';
/**
 * Creates a listing on a given asset and then executes a sell on the created bid and listing.
 *
 * ```ts
 * await metaplex
 *   .auctionHouse()
 *   .sell({ auctionHouse, bid })
 *   .run();
 * ```
 *
 * @group Operations
 * @category Constructors
 */

const directSellOperation = Operation.useOperation(Key);
/**
 * @group Operations
 * @category Types
 */

/**
 * @group Operations
 * @category Handlers
 */
const directSellOperationHandler = {
  handle: async (operation, metaplex) => (await directSellBuilder(metaplex, operation.input)).sendAndConfirm(metaplex, operation.input.confirmOptions)
}; // -----------------
// Builder
// -----------------

/**
 * @group Transaction Builders
 * @category Inputs
 */

/**
 * Creates a listing on a given asset and executes a sale on the created listing and given bid.
 *
 *
 * ```ts
 * const transactionBuilder = metaplex
 *   .auctionHouse()
 *   .builders()
 *   .sell({ auctionHouse, bid, seller })
 * ```
 *
 * @group Transaction Builders
 * @category Constructors
 */
const directSellBuilder = async (metaplex, params) => {
  var _params$printReceipt;

  // Data.
  const {
    auctionHouse,
    auctioneerAuthority,
    bid,
    seller = metaplex.identity(),
    authority = auctionHouse.authorityAddress,
    bookkeeper = metaplex.identity(),
    createListingInstructionKey,
    executeSaleInstructionKey
  } = params;
  const {
    hasAuctioneer
  } = auctionHouse;
  const {
    tokens,
    price,
    buyerAddress
  } = bid;
  const printReceipt = ((_params$printReceipt = params.printReceipt) !== null && _params$printReceipt !== void 0 ? _params$printReceipt : true) && Boolean(bid.receiptAddress);

  if (hasAuctioneer && !auctioneerAuthority) {
    throw new errors.AuctioneerAuthorityRequiredError();
  }

  const asset = Bid.isPrivateBid(bid) ? bid.asset : { ...bid.asset,
    token: params.sellerToken
  };
  const listingBuilder = await createListing.createListingBuilder(metaplex, {
    mintAccount: asset.mint.address,
    price: price,
    auctionHouse,
    auctioneerAuthority,
    seller,
    authority,
    tokenAccount: asset.token.address,
    tokens,
    printReceipt,
    bookkeeper,
    instructionKey: createListingInstructionKey
  });
  const {
    receipt,
    sellerTradeState
  } = listingBuilder.getContext();
  const listing = {
    model: 'listing',
    lazy: false,
    auctionHouse,
    asset,
    tradeStateAddress: sellerTradeState,
    bookkeeperAddress: PublicKey.toPublicKey(bookkeeper),
    sellerAddress: PublicKey.toPublicKey(seller),
    receiptAddress: receipt,
    purchaseReceiptAddress: null,
    price,
    tokens: tokens,
    createdAt: DateTime.now(),
    canceledAt: null
  };
  const saleBuilder = await executeSale.executeSaleBuilder(metaplex, {
    auctionHouse,
    auctioneerAuthority,
    bid,
    listing,
    printReceipt,
    bookkeeper,
    instructionKey: executeSaleInstructionKey
  });
  const {
    receipt: receiptAddress
  } = saleBuilder.getContext();
  const buyerTokenAccount = pdas.findAssociatedTokenAccountPda(asset.address, buyerAddress);
  const purchasedAsset = { ...asset,
    token: { ...asset.token,
      address: buyerTokenAccount,
      ownerAddress: buyerAddress
    }
  };
  const purchase = {
    auctionHouse,
    model: 'purchase',
    lazy: false,
    asset: purchasedAsset,
    buyerAddress,
    sellerAddress: PublicKey.toPublicKey(seller),
    bookkeeperAddress: PublicKey.toPublicKey(bookkeeper),
    receiptAddress,
    price: bid.price,
    tokens,
    createdAt: DateTime.now()
  };
  return TransactionBuilder.TransactionBuilder.make().setContext({
    listing,
    purchase
  }).add(listingBuilder).add(saleBuilder);
};

exports.directSellBuilder = directSellBuilder;
exports.directSellOperation = directSellOperation;
exports.directSellOperationHandler = directSellOperationHandler;
//# sourceMappingURL=directSell.cjs.map
